#
# Generic Conference
# ------------------

#
# This module contains routines which are common to all groupware conference
# applications.
#
# The generic conference essentially handles all the details of maintaining
# connections with other conference users; the main data structure for this
# is the "_gk_remoteUsers" list, which holds the list of all other users, where each
# user is represented as a keyed list.  Interactions between the generic
# conference and the Registrar Clients (see rc.tcl) result in users being
# added to or deleted from this _gk_remoteUsers list.
#
# Each keyed list in the "_gk_remoteUsers" list maintains at least the following
# useful information on each user:
#
#    usernum  - the unique user number for that user
#    userid   - the login id of the user
#    username - the user's proper name (if available) or the userid
#    filedesc - the file descriptor used by tcl-dp to send messages on
#
# Additionally, the variable "_gk_localUser" is a keyed list containing information
# about ourself, including at least the following:
#
#    confname   - the name of our conference
#    conftype   - the type of our conference
#    originator - the host and port of the Registrar Client who initiated
#                 the conference
#    confnum    - the unique id number of the conference
#    reghost    - the host of the Registrar Client that started the conference
#    regport    - the port of the Registar Client that started the conference
#    usernum    - our unique user number 
#    username   - our proper name (if available) or the userid
#    userid     - our login id
#

_gk_initGkErrorHandling

catch {source ~/.tclgkrc}
if {[info exists internetdomain] == 0} {
    set internetdomain ""
}
set hostprefix [exec hostname]
if {[string index $internetdomain 0] == "."} {
    set host $hostprefix$internetdomain
} else {
    if {$internetdomain == ""} {
        set host $hostprefix
    } else {
        set host $hostprefix.$internetdomain
    }
}

#
# This routine initializes the conference application, and should be the
# first thing called by the actual application conference.  It sets up
# several things, and the connects back to the Registar Client that spawned
# us, and then telling the Registar Client who we are so that messages from
# the registration system can be sent to us (see rc.tcl).
#

proc gk_initConf {argv} { global _gk_remoteUsers _gk_localUser host _gk_regClient
    set myport [dp_MakeRPCServer 0]
    set _gk_remoteUsers {}
    set _gk_localUser $argv
    set _gk_regClient [dp_MakeRPCClient [keylget _gk_localUser reghost] \
		   [gk_getLocalAttrib regport]]
    keylset msg confnum [gk_getLocalAttrib confnum] host $host \
	port $myport usernum [gk_getLocalAttrib usernum]
    dp_RDO $_gk_regClient _gk_remoteInfo $msg
}


#  Trace gk_newUser so that we can do something everytime there is an
#  entrant into an existing conference.
#  What do we want to do?  We want one of the existing conference
#  participants to update the entrant.  The participant with the
#  smallest usernum is chosen to do the updating.
#  NOTE:  In order to use this feature, applications must provide a
#  trace on gk_entrantUserNum.

trace variable gk_newUser w _gk_doUpdateEntrant
proc _gk_doUpdateEntrant args { global _gk_remoteUsers gk_newUser _gk_localUser gk_entrantUserNum
    set smallest [gk_getLocalAttrib usernum]
    foreach i $_gk_remoteUsers {
	if {$smallest > [keylget i usernum]} { return }
    }
    set gk_entrantUserNum [keylget gk_newUser usernum]
    return
}

#
# This routine is used to exchange information with other conferences
# that establish connections to us (usually as a result of _gk_connectTo
# calls initiated by the registration system).  We receive the remote user's
# user number and other information, and if requested, send the remote user
# our information.
#
# When we receive this information, the remote user is officially "in" our
# conference, and is added to the _gk_remoteUsers list automatically.  If you wish
# your program to be notified when new users enter, put a write trace on
# the global variable "gk_newUser" which will be set to the keyed list of
# information on the new user.
#
proc _gk_remoteInfo {whom echo} {  global rpcFile _gk_remoteUsers gk_newUser _gk_localUser
    set flag 0
    if {[_gk_findUser [keylget whom usernum]] == ""} {
	keylset whom filedesc $rpcFile
	lappend _gk_remoteUsers $whom
	set flag 1
    }
    if {$echo} {
	dp_RDO $rpcFile _gk_remoteInfo $_gk_localUser 0
    }
    if {$flag == 1} {
	set gk_newUser $whom
    }
}


#
# The registration system calls this to tell us to hook up to a particular
# user, whose host and port number are specified in the parameter whom.  
# After initiating a connection, we send the remote conference information
# on ourselves, and request that they send us their information.  Note that
# the new user is not in the conference until we receive this information
# about them (see proc _gk_remoteInfo, above).
#

proc _gk_connectTo whom { global _gk_localUser
    set file [dp_MakeRPCClient [keylget whom host] [keylget whom port]]
    dp_RDO $file _gk_remoteInfo $_gk_localUser 1
}


#
# The registration system calls this to tell us to remove the indicated user
# from the conference.  The _gk_remoteUsers list is updated.  If the user is ourself
# the _gk_deleteConf routine (below) is called to delete the conference application
#
# If you wish your program to be notified when users leave a conference,
# put a write trace on the global variable "gk_deletedUser" which will be set
# to a keyed list containing at least the user number of the deleted user.
#

proc _gk_removeUser whom { global _gk_remoteUsers gk_deletedUser _gk_localUser
    set i 0; while { $i < [llength $_gk_remoteUsers] } {
	set tmp [lindex $_gk_remoteUsers $i]
	if {[keylget tmp usernum] == [keylget whom usernum] } {
	    set _gk_remoteUsers [lreplace $_gk_remoteUsers $i $i]
	    set gk_deletedUser $whom
	}
	incr i
    }
    if {[keylget _gk_localUser usernum] == [keylget whom usernum]} {
	_gk_deleteConf
    }
}


#
# The procedure called when this conference is to be deleted, either 
# because the whole conference (ie. all users) are deleted, or because
# the local user has left the conference.  You may wish to override this
# to do some sort of cleanup.
#

proc _gk_deleteConf {} {
    exit
}


########################################
# This routine returns the keyed list representing the particular user
# given their user number.
#
proc _gk_findUser idnum {  global _gk_remoteUsers;
    foreach i $_gk_remoteUsers {
	set user [keylget i usernum]
	if {$user == $idnum} { return $i }
    }
    return ""
}
########################################

########################################
# Send the command in args to ALL users, including the local user.
proc gk_toAll args { global _gk_remoteUsers 
    uplevel #0 $args 
    foreach i $_gk_remoteUsers { dp_RDO [keylget i filedesc] eval $args }
}

# Send the command in args to REMOTE users only.
proc gk_toOthers args { global _gk_remoteUsers
    foreach i $_gk_remoteUsers { dp_RDO [keylget i filedesc] eval $args }
}

# Send the command in args to the user with usernum == UserNum.
proc gk_toUserNum {UserNum args}  {global _gk_remoteUsers
    dp_RDO [gk_getUserAttrib $UserNum filedesc] eval $args
}

# Get attribute AttribName from user with usernum == UserNum.
proc gk_getUserAttrib {UserNum AttribName} {global _gk_remoteUsers
    foreach i $_gk_remoteUsers {
       if {[keylget i usernum] ==  $UserNum} {break}
    }
    return [keylget i $AttribName]
}

# Set attribute AttribName for user with usernum == UserNum to NewValue.
proc gk_setUserAttrib {UserNum AttribName NewValue} {global _gk_remoteUsers
    foreach i $_gk_remoteUsers {
       if {[keylget i usernum] ==  $UserNum} {break}
    }
    keylset i $AttribName $NewValue
}
 
# Get attribute AttribName from local user.
proc gk_getLocalAttrib {AttribName} {global _gk_localUser
    return [keylget _gk_localUser $AttribName]
}

# Set attribute AttribName for local user to NewValue.
proc gk_setLocalAttrib {AttribName NewValue} {global _gk_localUser
    keylset _gk_localUser $AttribName $NewValue
}
 
# Returns 1 if this conference application is the conference
# originator, 0 otherwise.
proc gk_amOriginator {} {
    global _gk_localUser
    if {[gk_getLocalAttrib originator] == "[gk_getLocalAttrib reghost][gk_getLocalAttrib regport]"} {
	return 1
    }
    return 0
}
########################################
