#
# Generic Registrar Client
# ------------------------
#
# This is not a runnable program itself, but can be sourced by "specific"
# Registrar Client programs (e.g. openrc.tcl).
#
# A Registrar Client connects to the central Registrar (see reg.tcl), and
# sends requests to modify the lists of conferences and their users that the
# Registrar maintains.  The Registrar Client acts as the "controller" in
# an open protocol, implementing specific policies that determine how users 
# can join and leave conferences.  It is possible to define many such
# Registrar Clients, offering different semantics and interfaces.
#
# This file contains generic routines which interface with the Registrar,
# for sending and receiving messages, etc.  It is up to particular Registrar
# Clients to implement at least the following routines which are called by
# the generic Registrar Client code (see openrc.tcl) for one implementation:
#
#     gk_foundNewConf     - called when a new conference has been created
#     gk_foundNewUser     - called when a new user is found for a conference
#     gk_foundDeletedUser - called when a user is deleted from a conference
#     gk_foundDeletedConf - called when a conference is deleted
#
# This file also contains routines which spawn and then interact with
# particular groupware applications (termed conferences).  This subsumes
# the functionality of the Coordinator object from the InterViews 
# implementation.  These routines forward directives from the registration
# system (such as requests to connect to a new user) to the appropriate
# conference application.
#

proc _gk_initGenericRC {} {

}

_gk_initGkErrorHandling 

catch {source ~/.tclgkrc}

foreach i $argv {
    if {[string range $i 0 1] == "-p"} {
	set regportnum [string range $i 2 [expr [string length $i]-1]]
    }
    if {[string range $i 0 1] == "-h"} {
	set reghostname [string range $i 2 [expr [string length $i]-1]]
    }
}

if {[info exists myusername] == 0} {
    set myusername [exec whoami]
}
if {[info exists internetdomain] == 0} {
    set internetdomain ""
}
if {[info exists regportnum] == 0} {
    set regportnum 9068
}
if {[info exists reghostname] == 0} {
    set reghostname manaslu
}

#
# connect up to a particular central Registrar.  Change this to connect
#    to a different Registrar (e.g. for offsite conferences).  This will
#    eventually be expanded so that a command line flag can override this.
#

set registrar [dp_MakeRPCClient $reghostname $regportnum]

#
# set up a server so that programs can connect to us (typically the groupware
#    conferences that we spawn, but conceivably other Registrar Clients also)
#

set myport [dp_MakeRPCServer 0]
set hostprefix [exec hostname]
if {[string index $internetdomain 0] == "."} {
    set host $hostprefix$internetdomain
} else {
    if {$internetdomain == ""} {
	set host $hostprefix
    } else {
	set host $hostprefix.$internetdomain
    }
}


#
# _gk_rcConfsList and _gk_rcUsersList() essentially echo the lists maintained by the
#    central Registrar.  
#

set _gk_rcConfsList {}
set _gk_rcUsersList(0) {}


#
# This is a generic proc used to update both the conference lists and the
#    user lists.  Assume we want to update the list of conferences.  The 
#    scenario here is as follows:
#      
#       1. The central Registrar calls the Registrar Client's "_gk_confList"
#          proc, passing it the most up to date list of conferences.  
#          (This is invoked from the Registrar's _gk_dispConference proc,
#          typically because one of the Registrar Clients called
#          _gk_pollConferences to request the conference list).
#
#       2. Our _gk_confList proc calls _gk_updateList, passing parameters as follows:
#
#           list: $_gk_regConfsList              - the newly updated list 
#                                       (a list of keyed lists)
#           oldlist: _gk_rcConfsList         - the most recent local copy of the list
#           root: _gk_rcConfsList            - ditto (used to declare list as global)
#           key: confnum              - the key from each keyed list whose 
#                                       value is a unique id number identifying
#                                       the list element
#           newproc: gk_foundNewConf     - a proc to call if a new item is found
#           delproc: gk_foundDeletedConf - a proc to call if an item was deleted
#
#       3. The proc basically compares the new list with the old list.
#          If items appear in the new list but not the old list, the
#          newproc is called.  If items appear in the old list but not the
#          new list, the delproc is called.  
#
#       4. This routine does NOT change oldlist (e.g. _gk_rcConfsList) so that its 
#          contents reflect that of the new list.  Such changes can of course
#          be made by the newproc and delprocs if appropriate for the 
#          particular registration policy.  While this seems weird, its
#          actually an important part of what gives the open protocol its
#          flexibility.
#

proc _gk_updateList {list oldlist root key newproc delproc} {
    global $root
    set newlist {}
    set gonelist {}
    set lst [eval [set cmd "concat \$$oldlist"]]

    set i 0; while {$i < [llength $list]} {
	set x [lindex $list $i]
	set cnfnum [keylget x $key]
	set pattern "*\{$key $cnfnum\}*"
	set posn [lsearch $lst $pattern]
	if {$posn == -1} { lappend newlist $x }
	incr i
    }

    set i 0; while {$i < [llength $lst]} {
	set x [lindex $lst $i]
	set cnfnum [keylget x $key]
	set pattern "*\{$key $cnfnum\}*"
	set posn [lsearch $list $pattern]
	if {$posn == -1} { lappend gonelist $cnfnum }
	incr i
    }

    foreach i $newlist { eval $newproc [list $i] }
    foreach i $gonelist { eval $delproc [list $i] }
}


#
# update the conference list in response to new info from the registrar
#

proc _gk_confList _gk_regConfsList {
    _gk_updateList $_gk_regConfsList _gk_rcConfsList _gk_rcConfsList confnum gk_foundNewConf gk_foundDeletedConf
}


#
# update the _gk_rcUsersList for a particular conference in response to new info
#    from the registrar.  
#

proc _gk_userList {conf _gk_regUsersList} { 
    global _gk_rcUsersList
    if {[info exists _gk_rcUsersList($conf)] == 0} {
	set _gk_rcUsersList($conf) ""
    }
    set zzz "gk_foundDeletedUser $conf"
    _gk_updateList $_gk_regUsersList _gk_rcUsersList($conf) _gk_rcUsersList usernum gk_foundNewUser $zzz
}


#
# a utility proc that given a unique conference number will locate the
#    keyed list containing the conference's information and return it to
#    the caller

proc _gk_findConfInfo confnum {  global _gk_rcConfsList
    foreach i $_gk_rcConfsList {
	set conf [keylget i confnum]
	if {$conf == $confnum} { return $i }
    }
    return ""
}



#
# These routines are used to send messages to the central Registrar.  They
#    can be called by routines in the specific Registrar Client.  All are
#    just for convenience and reasonably straightforward.  See openrc.tcl
#    for examples of how they can be used
#

proc _gk_callJoinConference confnum {	global myport registrar host myusername
	keylset user confnum $confnum userid [exec whoami] host $host port $myport \
	    username $myusername
	dp_RDO $registrar _gk_addUser $user
}

proc _gk_callJoinConfWithKeys {user} { global registrar; dp_RDO $registrar _gk_addUser $user }
proc _gk_callLeaveConference {conf user} { global registrar; dp_RDO $registrar _gk_deleteUser $conf $user }
proc _gk_callNewConference conf { global registrar; dp_RDO $registrar _gk_newConference $conf }
proc _gk_callDeleteConference confnum { global registrar; dp_RDO $registrar _gk_deleteConference $confnum }
proc _gk_pollConferences {} { global registrar; dp_RDO $registrar _gk_dispConference }
proc _gk_pollUsers confnum { global registrar; dp_RDO $registrar _gk_dispUsers $confnum }


#
# The rest of this file deals with connections to the conferences that we
#    create - the so-called "Coordinator" responsibilities.  
#
# The first thing we keep is a list of all the conferences connected to us.
#    Each conference is represented as a keyed list, with information about
#    the conference.  This information is transmitted by the conference
#    after it has been created and connected back to us. 
#

set conferences ""


#
# Given a unique conference number, return the keyed list containing that
#    conference's information
#

proc _gk_findConf {confnum} {  global conferences
    foreach i $conferences {
	set conf [keylget i confnum]
	if {$conf == $confnum} { return $i }
    }
    return ""
}


#
# This routine creates a groupware application process (conference).  
#    The first parameter is a keyed list containing information about the
#    conference, while the second specifies the user number which will be
#    given the local user of the conference.
#
# This routine assumes the existence of an option database (ie. .Xdefaults)
# containing a block of entries like the following:
#
#    *conferenceTypes:       3
#    *conf1-desc:            Brainstorming Conference
#    *conf1-prog:            gkwish -f brainstorm.tcl
#    *conf2-desc:            GroupSketch
#    *conf2-prog:            gkwish -f sketch.tcl
#    *conf3-desc:            PostIt Notes
#    *conf3-prog:            gkwish -f postit.tcl
#
# The conferenceTypes resource specifies the number of different conferences
# which exist, while the remaining pairs of resources specify the type of the
# conference and the actual command used to start the conference application.
#
# This routine looks at the conftype key in the 'conf' parameter, and finds
# the conf?-desc resource which matches it.  The corresponding conf?-prog
# resource gives the command line of the program to execute.  This is
# supplemented with information needed by the conference application, so that
# the resulting information includes (at a minimum):
#
#    confname   - the name assigned to the conference
#    conftype   - the type of the conference (used above to run the program)
#    originator - the host and port number (e.g. janu9865) of the Registrar
#                 client which first created the conference
#    confnum    - the unique conference id number assigned by the Registrar 
#    reghost    - the host of the Registrar Client (i.e. us), which the new
#                 conference will connect to as soon as it is created
#    regport    - the port number of the Registrar Client (i.e. us)
#    usernum    - the unique user number assigned by the Registrar
#    userid     - the userid (login) of the user
#
# The complete command line (name of program plus above information) is then
# exec'ed to start up the actual conference application.
#
# NOTE: we need to put a timer on this and check to make sure the conference
#       really did get created and connect back up; if not, delete it via
#       the Registrar
#
# CHANGED to use gk_program variable rather than Xdefaults

proc _gk_createConference {conf usernum} { global myport myusername host gk_program
#     global _gk_stillCheck; set _gk_stillCheck yes
     global _gk_stillCheck
     set _gk_stillCheck($usernum) "yes"

     after 10000 "_gk_checkConnection $conf $usernum"
     set cmdline $gk_program([keylget conf conftype])
     keylset conf reghost $host regport $myport usernum $usernum \
        userid [exec whoami] username $myusername
     set cmd [concat $cmdline " " $conf " &"]
     eval $cmd
}

# This is called to check whether the conference has not connected  back
# to the registrar client within three seconds.  If it has not, it is
# removed.
proc _gk_checkConnection args { global _gk_stillCheck
    set conf [lindex [lindex $args 3] 1]
    set user [lindex $args 4]
    if {$_gk_stillCheck($user) == "no"} {return}
    if {[_gk_findConf $conf] == ""} {
	_gk_userLeft $conf $user no
    }
}

#
# One of the first things that the new conference will do is connect back
#    up to us and call this method, which essentially tells us which conference
#    is hooked up (so we can associate the file description with the conference
#    number.  This routine also will send any pending messages to the 
#    newly connected conference (see the _gk_toConf routine, immediately below).
#

proc _gk_remoteInfo whom {  global conferences rpcFile pending
    keylset whom filedesc $rpcFile
    lappend conferences $whom
    if {[info exists pending([keylget whom confnum])]} {
	foreach i $pending([keylget whom confnum]) {
	    eval [concat dp_RDO [keylget whom filedesc] $i]
	}
    } 
}


#
# This routine is used to send messages to a particular conference.  This
#    routine is used rather than sending directly, because it may be desirable
#    to send messages to the conference *before* its actually been created
#    and connected back to us (remember there is a time lag between when we
#    ask it to be created and it really is there).  If the conference is
#    present and can accept messages, the message is send immediately.  If
#    not, the message is queued up in the pending() array, whose contents
#    will be sent to the conference when it does hook up, via the 
#    _gk_remoteInfo method above.
proc _gk_toConf {confnum msg} { global pending conferences
    set conf [_gk_findConf $confnum]
    if {$conf == ""} {
	lappend pending($confnum) $msg
    } else {
	eval [concat dp_RDO [keylget conf filedesc] $msg]
    }
}


#
# This routine is used to tell the local instantiation of a particular
#    conference to connect up to the indicated user (e.g. because that
#    user just joined the conference.
#
# Things here are not quite as straightforward as they seem.  While the
# 'user' parameter does contain a host and a port number, this host and
# port number is *not* the host and port of the conference application
# that our conference should connect to, but in fact the host and port
# number of that user's Registrar Client.  (There's a reason for this,
# but don't worry about it).
#
# Therefore, we have to first locate the actual host and port number of
# the user's conference application.  This is done (surprise) through the
# user's Registrar Client, in particular the _gk_address routine (below).  
# After retrieving the correct host and port number, we then instruct
# our conference to connect to the new user's conference application.
#
# We do some retry things in case the remote conference hasn't been created
# yet.  Not well tested yet.
#
proc _gk_joinTo {user conf} {
    set remote [dp_MakeRPCClient [keylget user host] [keylget user port]]
    set addr [dp_RPC $remote _gk_address [keylget conf confnum]]
    dp_CloseRPC $remote
    if {$addr == ""} {
	after 1000 _gk_joinTo [list $user] [list $conf]
    } else {
	_gk_toConf [keylget conf confnum] [concat _gk_connectTo \{$addr\}]
    }
}


#
# Given the unique conference number, return the host and port number of 
#    the corresponding conference.  Used by the above routine.
#

proc _gk_address confnum {
    set theconf [_gk_findConf $confnum]
    if {$theconf == ""} {
	return ""
    } else {
	return $theconf
    }
}


#
# This is called (via routines in errors.tcl) if a socket connection closes.
#    Presumably the other end of the socket was one of the groupware 
#    conference applications we spawned.  Call the "_gk_userLeft" routine (below)
#    to do some clean up.
#

proc socket_closed filedesc {  global conferences
    set i 0; while { $i < [llength $conferences] } {
	set tmp [lindex $conferences $i]
	if {[keylget tmp filedesc] == $filedesc} {
	    _gk_userLeft [keylget tmp confnum] [keylget tmp usernum] $filedesc
	}
	incr i
    }
}


#
# After a conference is killed (gracefully or otherwise), first tell the
#    central Registrar that the user has left the conference.  If the user
#    was the last user in the conference, tell the Registrar to delete the
#    conference, since no one is left.
#

proc _gk_userLeft {confid userid index} {  
     global _gk_rcUsersList conferences  _gk_stillCheck
#     set _gk_stillCheck no
     set _gk_stillCheck($userid) "no"
    if {$index != "no"} {
	for {set i 0} {$i < [llength $conferences]} {incr i 1} {
	    set tmp [lindex $conferences $i]
	    if {[keylget tmp filedesc] == $index} {
		set conferences [lreplace $conferences $i $i]
	    }
	}
    }
    _gk_callLeaveConference $confid $userid
    if {[llength $_gk_rcUsersList($confid)] < 2} {
	if {[info exists _gk_rcUsersList] == 0} {return}
	_gk_callDeleteConference $confid
    }
    _gk_pollUsers $confid
    _gk_pollConferences
}
